"""
Originally from Evan Walters and Omead Pooladzandi, 2024
Modified under Creative Commons Attribution 4.0 International
Source available at https://github.com/evanatyourservice/kron_torch/blob/97a2b5ee8a1a4c29e4780bbf6c521e545189eff9/kron_torch/kron.py
"""

import torch
from heavyball.utils import copy_stochastic_list_

from .utils import update_param_, warmup, psgd_precond_grad, init_Q_exprs, trust_region_clip_, PSGDBase, \
    precond_update_prob_schedule, split_p_and_g_in_group, triu_to_line, line_to_triu, set_, promote


class ForeachDelayedPSGD(PSGDBase):
    """
    Implements PSGD with off-by-one preconditioning (akin to ADOPT and SOAP)

    Args:
        params (iterable): Iterable of parameters to optimize or dicts defining
            parameter groups.
        lr (float): Learning rate.
        b1 (float): Momentum parameter.
        weight_decay (float): Weight decay (L2 penalty).
        preconditioner_update_probability (callable or float, optional): Probability of
            updating the preconditioner. If None, defaults to a schedule that anneals
            from 1.0 to 0.03 by 4000 steps.
        max_size_triangular (int): Max size for dim's preconditioner to be triangular.
        min_ndim_triangular (int): Minimum number of dimensions a layer needs
            to have triangular preconditioners.
        memory_save_mode: (string, optional), None, 'one_diag', or 'all_diag', None is default
            to set all preconditioners to be triangular, 'one_diag' sets the largest
            or last dim to be diagonal per layer, and 'all_diag' sets all preconditioners
            to be diagonal.
        momentum_into_precond_update: (bool), whether to send momentum into preconditioner
            update instead of raw gradients.
    """

    def __init__(self, params, lr=0.001, beta=0.9, weight_decay=0.0, preconditioner_update_probability=None,
                 max_size_triangular=2048, min_ndim_triangular=2, memory_save_mode=None,
                 momentum_into_precond_update=True, warmup_steps: int = 1, merge_dims: bool = False,
                 split: bool = False, clip_fn: callable = None, store_triu_as_line: bool = True,
                 foreach: bool = True, q_dtype='float32', stochastic_schedule: bool = True,  #
                 # expert parameters
                 precond_init_scale=1.0, precond_lr=0.1):
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= beta < 1.0:
            raise ValueError(f"Invalid beta parameter: {beta}")
        if not 0.0 <= weight_decay:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")

        if clip_fn is None:
            clip_fn = lambda x: trust_region_clip_(x, 0.9, 1.5)

        defaults = dict(lr=lr, beta=beta, weight_decay=weight_decay, max_size_triangular=max_size_triangular,
                        min_ndim_triangular=min_ndim_triangular, memory_save_mode=memory_save_mode,
                        momentum_into_precond_update=momentum_into_precond_update, precond_lr=precond_lr,
                        precond_init_scale=precond_init_scale,
                        step=0, warmup_steps=warmup_steps, merge_dims=merge_dims, split=split,
                        store_triu_as_line=store_triu_as_line, q_dtype=q_dtype)
        super().__init__(params, defaults, foreach, stochastic_schedule, clip_fn, preconditioner_update_probability)


    def _step(self, group):
        momentum_into_precond_update = group.get("momentum_into_precond_update", True)
        precond_init_scale = group['precond_init_scale']
        max_size_triangular = group['max_size_triangular']
        min_ndim_triangular = group['min_ndim_triangular']
        memory_save_mode = group['memory_save_mode']
        precond_lr = group['precond_lr']
        weight_decay = group['weight_decay']
        lr = group['lr']
        beta = group['beta']
        store_triu_as_line = group['store_triu_as_line']
        q_dtype = getattr(torch, group['q_dtype'])

        vals = []

        for p, g in split_p_and_g_in_group(group):
            state = self.state_(p)

            if 'Q' not in state:
                state["exp_avg"] = torch.zeros_like(g)
                Q, state["exprs"] = init_Q_exprs(p, precond_init_scale, max_size_triangular, min_ndim_triangular,
                                                 memory_save_mode, dtype=q_dtype)
                state["Q"] = triu_to_line(Q) if store_triu_as_line else Q

            vals.append((p, g, state["exp_avg"], state["Q"]))

        if not vals:
            return

        p_list, grad_list, exp_avg_list, Q_list = zip(*vals)
        del vals

        group["step"] += 1

        torch._foreach_lerp_(exp_avg_list, grad_list, (1 - beta) / (1 - beta ** group["step"]))

        Q_list, exp_avg_list = list(Q_list), list(exp_avg_list)
        for i, (p, g) in enumerate(zip(p_list, grad_list)):
            q_orig = Q_list.pop(0)
            ea = exp_avg_list.pop(0)
            q = line_to_triu(q_orig) if store_triu_as_line else q_orig
            new = psgd_precond_grad(q, self.state_(p)["exprs"], ea)
            if self.should_update(group):
                q32 = [promote(q_) for q_ in q]
                self.do_update(group,[p], [ea if momentum_into_precond_update else g], [q32], precond_lr, [q_orig], store_triu_as_line)
            set_(g, new)

        grad_list = self.clip_fn(grad_list)

        lr = -warmup(lr, group['step'], group['warmup_steps'])
        update_param_(p_list, grad_list, lr, weight_decay)



from .cached_psgd_kron import ForeachCachedPSGDKron
from .delayed_psgd import ForeachDelayedPSGD
from .foreach_adamw import ForeachAdamW
from .foreach_adopt import ForeachADOPT
from .foreach_laprop import ForeachLaProp
from .foreach_sfadamw import ForeachSFAdamW
from .foreach_soap import ForeachSOAP
from .p_adam import ForeachPaLMPAdam
from .palm_foreach_sfadamw import PaLMForeachSFAdamW
from .palm_foreach_soap import PaLMForeachSOAP
from .precond_schedule_foreach_soap import PrecondScheduleForeachSOAP
from .precond_schedule_palm_foreach_soap import PrecondSchedulePaLMForeachSOAP
from .precond_schedule_sfpsoap import PrecondScheduleSFPaLMSOAP
from .psgd_kron import ForeachPSGDKron
from .pure_psgd import ForeachPurePSGD
from .schedule_free_palm_foreach_soap import SFPaLMForeachSOAP
from .cached_delayed_psgd_kron import ForeachCachedDelayedPSGDKron

PalmForEachSoap = PaLMForeachSOAP

PaLMSOAP = PaLMForeachSOAP
PaLMSFAdamW = PaLMForeachSFAdamW
PaLMSFSoap = SFPaLMForeachSOAP
PrecondScheduleSFPaLMSOAP = PrecondScheduleSFPaLMSOAP
SOAP = ForeachSOAP
SFAdamW = ForeachSFAdamW
LaProp = ForeachLaProp
ADOPT = ForeachADOPT
PrecondScheduleSOAP = PrecondScheduleForeachSOAP
PrecondSchedulePaLMSOAP = PrecondSchedulePaLMForeachSOAP
PSGDKron = ForeachPSGDKron
AdamW = ForeachAdamW
PurePSGD = ForeachPurePSGD
PaLMPAdam = ForeachPaLMPAdam
DelayedPSGD = ForeachDelayedPSGD
CachedPSGDKron = ForeachCachedPSGDKron
CachedDelayedPSGDKron = ForeachCachedDelayedPSGDKron

__all__ = ['PalmForEachSoap', 'PaLMForeachSFAdamW', 'PaLMForeachSOAP', 'SFPaLMForeachSOAP', 'PrecondScheduleSFPaLMSOAP',
           'ForeachSOAP', 'ForeachSFAdamW', 'ForeachLaProp', 'ForeachADOPT', 'PrecondScheduleForeachSOAP',
           'PrecondSchedulePaLMForeachSOAP', 'ForeachPSGDKron', 'ForeachAdamW', 'ForeachPurePSGD', 'ForeachPaLMPAdam',
           'ForeachDelayedPSGD', 'ForeachCachedPSGDKron', 'ForeachCachedDelayedPSGDKron',  #
           'PaLMSOAP', 'PaLMSFAdamW', 'PaLMSFSoap', 'PaLMSFAdamW', 'PrecondScheduleSFPaLMSOAP',
           'SOAP', 'SFAdamW', 'LaProp', 'ADOPT', 'PSGDKron', 'AdamW', 'PurePSGD', 'PaLMPAdam', 'DelayedPSGD',
           'CachedPSGDKron', 'CachedDelayedPSGDKron', 'PrecondScheduleSOAP', 'PrecondSchedulePaLMSOAP']



from itertools import chain

import pytest
import torch

from heavyball import utils
from heavyball.utils import dim_merger, promote


def init_preconditioner(grad, state, precondition_frequency=10, shampoo_beta=0.95, max_precond_dim=10000,
                        precondition_1d=False, merge_dims=False):
    """
    Initializes the preconditioner matrices (L and R in the paper).
    """
    state['GG'] = []  # Will hold all the preconditioner matrices (L and R in the paper).
    if grad.dim() == 1:
        if not precondition_1d or grad.shape[0] > max_precond_dim:
            state['GG'].append([])
        else:
            state['GG'].append(torch.zeros(grad.shape[0], grad.shape[0], device=grad.device, dtype=grad.dtype))
    else:
        if merge_dims:
            grad = dim_merger(grad, max_precond_dim)

        for sh in grad.shape:
            if sh > max_precond_dim:
                state['GG'].append([])
            else:
                state['GG'].append(torch.zeros(sh, sh, device=grad.device, dtype=grad.dtype))

    state['Q'] = None  # Will hold all the eigenbases of the preconditioner.
    state['precondition_frequency'] = precondition_frequency
    state['shampoo_beta'] = shampoo_beta


def project(grad, state, merge_dims=False, max_precond_dim=10000):
    """
    Projects the gradient to the eigenbases of the preconditioner.
    """
    original_shape = grad.shape
    if merge_dims:
        grad = dim_merger(grad, max_precond_dim)

    for mat in state['Q']:
        if len(mat) > 0:
            grad = torch.tensordot(grad, mat, dims=[[0], [0]], )
        else:
            permute_order = list(range(1, len(grad.shape))) + [0]
            grad = grad.permute(permute_order)

    if merge_dims:
        grad = grad.reshape(original_shape)
    return grad


def update_preconditioner(grad, state, max_precond_dim=10000, merge_dims=False, precondition_1d=False):
    """
    Updates the preconditioner matrices and the eigenbases (L, R, Q_L, Q_R in the paper).
    """
    if grad.dim() == 1:
        if precondition_1d and grad.shape[0] <= max_precond_dim:
            state['GG'][0].lerp_(grad.unsqueeze(1) @ grad.unsqueeze(0), 1 - state['shampoo_beta'])
    else:
        if merge_dims:
            new_grad = dim_merger(grad, max_precond_dim)
            for idx, sh in enumerate(new_grad.shape):
                if sh <= max_precond_dim:
                    outer_product = torch.tensordot(new_grad, new_grad, dims=[[*chain(range(idx), range(idx + 1,
                                                                                                        len(new_grad.shape)))]] * 2, )
                    state['GG'][idx].lerp_(outer_product, 1 - state['shampoo_beta'])
        else:
            for idx, sh in enumerate(grad.shape):
                if sh <= max_precond_dim:
                    outer_product = torch.tensordot(grad, grad,  # Contracts across all dimensions except for k.
                                                    dims=[[*chain(range(idx), range(idx + 1, len(grad.shape)))]] * 2, )
                    state['GG'][idx].lerp_(outer_product, 1 - state['shampoo_beta'])

    if state['Q'] is None:
        state['Q'] = get_orthogonal_matrix(state['GG'])
    if state['step'] > 0 and state['step'] % state['precondition_frequency'] == 0:
        state['Q'] = get_orthogonal_matrix_QR(state, max_precond_dim, merge_dims)


def project_back(grad, state, merge_dims=False, max_precond_dim=10000):
    """
    Projects the gradient back to the original space.
    """
    original_shape = grad.shape
    if merge_dims:
        grad = dim_merger(grad, max_precond_dim)
    for mat in state['Q']:
        if len(mat) > 0:
            grad = torch.tensordot(grad, mat, dims=[[0], [1]], )
        else:
            permute_order = list(range(1, len(grad.shape))) + [0]
            grad = grad.permute(permute_order)

    if merge_dims:
        grad = grad.reshape(original_shape)
    return grad


def get_orthogonal_matrix(mat):
    """
    Computes the eigenbases of the preconditioner using torch.linalg.eigh decomposition.
    """
    matrix = []
    for m in mat:
        if len(m) == 0:
            matrix.append([])
            continue
        if m.data.dtype in (torch.float16, torch.float32):
            float_data = False
            original_type = m.data.dtype
            original_device = m.data.device
            matrix.append(promote(m.data))
        else:
            float_data = True
            matrix.append(m.data)

    final = []
    for m in matrix:
        if len(m) == 0:
            final.append([])
            continue
        try:
            _, Q = torch.linalg.eigh(m + 1e-30 * torch.eye(m.shape[0], device=m.device))
        except:
            _, Q = torch.linalg.eigh(m.to(torch.float64) + 1e-30 * torch.eye(m.shape[0], device=m.device))
            Q = Q.to(m.dtype)
        Q = torch.flip(Q, [1])

        if not float_data:
            Q = Q.to(original_device).type(original_type)
        final.append(Q)
    return final


def get_orthogonal_matrix_QR(state, max_precond_dim=10000, merge_dims=False):
    """
    Computes the eigenbases of the preconditioner using one round of power iteration
    followed by torch.linalg.qr decomposition.
    """
    precond_list = state['GG']
    orth_list = state['Q']

    matrix = []
    orth_matrix = []
    for m, o in zip(precond_list, orth_list):
        if len(m) == 0:
            matrix.append([])
            orth_matrix.append([])
            continue
        if m.data.dtype in (torch.float16, torch.float32):
            float_data = False
            original_type = m.data.dtype
            original_device = m.data.device
            matrix.append(promote(m.data))
            orth_matrix.append(promote(o.data))
        else:
            float_data = True
            matrix.append(promote(m.data))
            orth_matrix.append(promote(o.data))

    orig_shape = state['exp_avg_sq'].shape
    if merge_dims:
        exp_avg_sq = dim_merger(state['exp_avg_sq'], max_precond_dim)
    else:
        exp_avg_sq = state['exp_avg_sq']

    final = []
    for ind, (m, o) in enumerate(zip(matrix, orth_matrix)):
        if len(m) == 0:
            final.append([])
            continue
        est_eig = torch.diag(o.T @ m @ o)
        sort_idx = torch.argsort(est_eig, descending=True)
        exp_avg_sq = exp_avg_sq.index_select(ind, sort_idx)
        o = o[:, sort_idx]
        power_iter = m @ o
        Q, _ = torch.linalg.qr(power_iter)

        if not float_data:
            Q = Q.to(original_device).type(original_type)
        final.append(Q)

    if merge_dims:
        exp_avg_sq = exp_avg_sq.reshape(orig_shape)

    state['exp_avg_sq'] = exp_avg_sq
    return final


def _init(size, max_precond, merge_dims, precondition_1d, beta, precondition_frequency=1):
    grad = torch.randn(size, dtype=torch.double)
    ref_state = {'step': 1, 'exp_avg': torch.randn_like(grad), 'exp_avg_sq': torch.randn_like(grad)}
    new_state = {'step': 1, **{k: v.clone() for k, v in ref_state.items() if isinstance(v, torch.Tensor)}}
    init_preconditioner(grad.clone(), ref_state, precondition_frequency=precondition_frequency, shampoo_beta=beta,
                        max_precond_dim=max_precond, precondition_1d=precondition_1d, merge_dims=merge_dims)
    utils.init_preconditioner(grad.clone(), new_state, max_precond_dim=max_precond, precondition_1d=precondition_1d,
                              merge_dims=merge_dims)
    return grad, ref_state, new_state


def _updated(size, max_precond, merge_dims, precondition_1d, beta, iterations, precondition_frequency=1):
    grad, ref_state, new_state = _init(size, max_precond, merge_dims, precondition_1d, beta, precondition_frequency)
    for _ in range(iterations):
        ref_state['step'] += 1
        new_state['step'] += 1
        grad = torch.randn_like(grad)
        update_preconditioner(grad.clone(), ref_state, max_precond, merge_dims, precondition_1d)
        utils.update_preconditioner(grad.clone(), new_state, max_precond, merge_dims, precondition_1d, beta,
                                    precondition_frequency == 1)
        yield grad, ref_state, new_state


def _check(ref, new):
    for k, rr in ref.items():
        if k not in new:
            continue
        nn = new[k]
        if isinstance(rr, list):
            for r, n in zip(rr, nn):
                if isinstance(r, torch.Tensor):
                    assert ref['step'] and k and torch.allclose(r, n)
        elif isinstance(rr, torch.Tensor):
            assert ref['step'] and k and torch.allclose(rr, nn)


_size = 16


@pytest.mark.parametrize('size', [(_size,), (_size,) * 2, (_size,) * 3])
@pytest.mark.parametrize('max_precond', [_size ** 2 * 2, _size * 2, _size // 2])
@pytest.mark.parametrize('merge_dims', [True, False])
@pytest.mark.parametrize('precondition_1d', [True, False])
@pytest.mark.parametrize('beta', [0.5, 0.9, 0.99])
@torch.no_grad()
def test_init(size, max_precond, merge_dims, precondition_1d, beta):
    grad, ref_state, new_state = _init(size, max_precond, merge_dims, precondition_1d, beta)
    _check(ref_state, new_state)


@pytest.mark.parametrize('size', [(_size,), (_size,) * 2, (_size,) * 3])
@pytest.mark.parametrize('max_precond', [_size ** 2 * 2, _size * 2, _size // 2])
@pytest.mark.parametrize('merge_dims', [True, False])
@pytest.mark.parametrize('precondition_1d', [True, False])
@pytest.mark.parametrize('beta', [0.5, 0.9, 0.99])
@torch.no_grad()
def test_ggt(size, max_precond, merge_dims, precondition_1d, beta, iterations: int = 5):
    for grad, ref_state, new_state in _updated(size, max_precond, merge_dims, precondition_1d, beta, iterations,
                                               precondition_frequency=10**12):
        _check(ref_state, new_state)


@pytest.mark.parametrize('size', [(_size,), (_size,) * 2, (_size,) * 3])
@pytest.mark.parametrize('max_precond', [_size ** 2 * 2, _size * 2, _size // 2])
@pytest.mark.parametrize('merge_dims', [True, False])
@pytest.mark.parametrize('precondition_1d', [True, False])
@pytest.mark.parametrize('beta', [0.5, 0.9, 0.99])
@torch.no_grad()
def test_update(size, max_precond, merge_dims, precondition_1d, beta, iterations: int = 5):
    for grad, ref_state, new_state in _updated(size, max_precond, merge_dims, precondition_1d, beta, iterations):
        _check(ref_state, new_state)


@pytest.mark.parametrize('size', [(_size,), (_size,) * 2, (_size,) * 3])
@pytest.mark.parametrize('max_precond', [_size ** 2 * 2, _size * 2, _size // 2])
@pytest.mark.parametrize('merge_dims', [True, False])
@pytest.mark.parametrize('precondition_1d', [True, False])
@pytest.mark.parametrize('beta', [0.5, 0.9, 0.99])
@pytest.mark.parametrize('back', [True, False])
@torch.no_grad()
def test_project(size, max_precond, merge_dims, precondition_1d, beta, back, iterations: int = 5):
    for grad, ref_state, new_state in _updated(size, max_precond, merge_dims, precondition_1d, beta, iterations):
        proj_ref = (project_back if back else project)(grad.clone(), ref_state, merge_dims, max_precond)
        proj_new = utils.project(grad.clone(), ref_state['Q'], merge_dims, max_precond, back)

        assert ref_state['step'] and torch.allclose(proj_ref.contiguous(), proj_new.contiguous())